@@ -32,8 +32,6 @@
 #include <openssl/x509.h>  /* work around OpenSSL bug: missing definition of STACK_OF */
 #include <openssl/tls1.h>
 
-#include <sys/stat.h>
-
 #include <ctype.h>
 #include <limits.h>
 #include <stdio.h>
@@ -180,44 +178,30 @@ int parse_min_tls(const char *str)
  */
 int load_config(struct vpn_config *cfg, const char *filename)
 {
-	int ret = ERR_CFG_UNKNOWN;
+	int ret = 0;
 	FILE *file;
-	struct stat stat;
-	char *buffer, *line, *saveptr = NULL;
+	char *line = NULL;
+	size_t len = 0;
+	ssize_t read;
 
 	file = fopen(filename, "r");
-	if (file == NULL)
-		return ERR_CFG_SEE_ERRNO;
-
-	if (fstat(fileno(file), &stat) == -1) {
+	if (file == NULL) {
 		ret = ERR_CFG_SEE_ERRNO;
-		goto err_close;
+		goto err_return;
 	}
-	if (stat.st_size == 0) {
-		ret = ERR_CFG_EMPTY_FILE;
-		goto err_close;
-	}
-
-	buffer = malloc(stat.st_size + 1);
-	if (buffer == NULL) {
-		ret = ERR_CFG_NO_MEM;
-		goto err_close;
-	}
-
-	// Copy all file contents at once
-	if (fread(buffer, stat.st_size, 1, file) != 1) {
-		ret = ERR_CFG_CANNOT_READ;
-		goto err_free;
-	}
-
-	buffer[stat.st_size] = '\0';
 
 	// Read line by line
-	for (line = strtok_r(buffer, "\n", &saveptr); line != NULL;
-	     line = strtok_r(NULL, "\n", &saveptr)) {
+	while ((read = getline(&line, &len, file)) != -1) {
 		char *key, *equals, *val;
 		int i;
 
+		// Ignore blank lines. We could argue that the string must be at least
+		// 3 chars to be valid, eg. 'x=\n' but let the rest of the function
+		// logic handle that. NOTE: getline includes the '\n' in the string,
+		// which is removed later on.
+		if (read < 2)
+			continue;
+
 		if (line[0] == '#')
 			continue;
 
@@ -254,14 +238,14 @@ int load_config(struct vpn_config *cfg, const char *filename)
 			strncpy(cfg->gateway_host, val, GATEWAY_HOST_SIZE);
 			cfg->gateway_host[GATEWAY_HOST_SIZE] = '\0';
 		} else if (strcmp(key, "port") == 0) {
-			unsigned long port = strtoul(val, NULL, 0);
+			long port = strtol(val, NULL, 0);
 
-			if (port == 0 || port > 65535) {
-				log_warn("Bad port in configuration file: \"%lu\".\n",
+			if (port < 1 || port > 65535) {
+				log_warn("Bad port in configuration file: \"%ld\".\n",
 				         port);
 				continue;
 			}
-			cfg->gateway_port = port;
+			cfg->gateway_port = (uint16_t)port;
 		} else if (strcmp(key, "username") == 0) {
 			strncpy(cfg->username, val, USERNAME_SIZE);
 			cfg->username[USERNAME_SIZE] = '\0';
@@ -463,15 +447,14 @@ int load_config(struct vpn_config *cfg, const char *filename)
 			goto err_free;
 		}
 	}
-
-	ret = 0;
+	if (ferror(file))
+		ret = ERR_CFG_SEE_ERRNO;
 
 err_free:
-	free(buffer);
-err_close:
+	free(line);
 	if (fclose(file))
 		log_warn("Could not close %s (%s).\n", filename, strerror(errno));
-
+err_return:
 	return ret;
 }
 
