@@ -29,12 +29,15 @@
 #include "tunnel.h"
 #include "http.h"
 #include "log.h"
+#include "userinput.h"
 #ifndef HAVE_X509_CHECK_HOST
 #include "openssl_hostname_validation.h"
 #endif
 
 #include <openssl/err.h>
+#ifndef OPENSSL_NO_ENGINE
 #include <openssl/engine.h>
+#endif
 #include <openssl/ui.h>
 #include <openssl/x509v3.h>
 #if HAVE_SYSTEMD
@@ -45,7 +48,6 @@
 #include <arpa/inet.h>
 #include <fcntl.h>
 #include <ifaddrs.h>
-#include <net/if.h>
 #include <netdb.h>
 #if HAVE_PTY_H
 #include <pty.h>
@@ -54,6 +56,7 @@
 #endif
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <netinet/tcp.h>
 #include <sys/wait.h>
 #include <sys/ioctl.h>
 #include <termios.h>
@@ -64,6 +67,7 @@
 #include <errno.h>
 #include <signal.h>
 #include <string.h>
+#include <assert.h>
 
 
 struct ofv_varr {
@@ -231,8 +235,8 @@ static int pppd_run(struct tunnel *tunnel)
 		} else {
 			static const char *const v[] = {
 				ppp_path,
-				"115200", // speed
-				":192.0.2.1", // <local_IP_address>:<remote_IP_address>
+				"230400", // speed
+				":169.254.2.1", // <local_IP_address>:<remote_IP_address>
 				"noipdefault",
 				"noaccomp",
 				"noauth",
@@ -324,6 +328,7 @@ static int pppd_run(struct tunnel *tunnel)
 
 		if (close(tunnel->ssl_socket))
 			log_warn("Could not close ssl socket (%s).\n", strerror(errno));
+		tunnel->ssl_socket = -1;
 		execv(pppd_args.data[0], (char *const *)pppd_args.data);
 		free(pppd_args.data);
 
@@ -355,7 +360,8 @@ static int pppd_run(struct tunnel *tunnel)
 	return 0;
 }
 
-static const char * const pppd_message[] = {
+static const char * const ppp_message[] = {
+#if HAVE_USR_SBIN_PPPD // pppd(8) - https://ppp.samba.org/pppd.html
 	"Has detached, or otherwise the connection was successfully established and terminated at the peer's request.",
 	"An immediately fatal error of some kind occurred, such as an essential system call failing, or running out of virtual memory.",
 	"An error was detected in processing the options given, such as two mutually exclusive options being used.",
@@ -376,6 +382,31 @@ static const char * const pppd_message[] = {
 	"The PPP negotiation failed because serial loopback was detected.",
 	"The init script failed (returned a non-zero exit status).",
 	"We failed to authenticate ourselves to the peer."
+#else // sysexits(3) - https://www.freebsd.org/cgi/man.cgi?query=sysexits
+	// EX_NORMAL = EX_OK (0)
+	"Successful exit.",
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, // 1-9
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,	// 10-19
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, // 20-29
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, // 30-39
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, // 40-49
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, // 50-59
+	NULL, NULL, NULL, NULL, // 60-63
+	// EX_USAGE (64)
+	"The command was used incorrectly, e.g., with the wrong number of arguments, a bad flag, a bad syntax in a parameter, or whatever.",
+	NULL, NULL, NULL, NULL, // 65-68
+	// EX_UNAVAILABLE (69)
+	"A service is unavailable. This can occur if a support program or file does not exist.",
+	// EX_ERRDEAD = EX_SOFTWARE (70)
+	"An internal software error has been detected.",
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, // 71-77
+	// EX_CONFIG (78)
+	"Something was found in an unconfigured or misconfigured state.",
+	NULL, // 79
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, // 80-89
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, // 90-98
+	NULL // EX_TERMINATE (99), PPP internal pseudo-code
+#endif
 };
 
 static int pppd_terminate(struct tunnel *tunnel)
@@ -387,56 +418,79 @@ static int pppd_terminate(struct tunnel *tunnel)
 
 	int status;
 
+	/*
+	 * Errors outside of the PPP process are returned as negative integers.
+	 */
 	if (waitpid(tunnel->pppd_pid, &status, 0) == -1) {
 		log_error("waitpid: %s\n", strerror(errno));
-		return 1;
+		return -1;
 	}
+
+	/*
+	 * Errors in the PPP process are returned as positive integers.
+	 */
 	if (WIFEXITED(status)) {
 		int exit_status = WEXITSTATUS(status);
 
+		/*
+		 * PPP exit status codes are positive integers. The way we interpret
+		 * their value is not straightforward:
+		 * - in the case of "normal" exit, the PPP process may return 0,
+		 *   but also a strictly positive integer such as 16 in the case of
+		 *   pppd,
+		 * - in the case of failure, the PPP process will return a strictly
+		 *   positive integer.
+		 * For now we process PPP exit status codes as follows:
+		 * - exit status codes synonym of success are logged and then
+		 *   translated to 0 before they are returned to the calling function,
+		 * - other exit status codes are considered synonyms of failure and
+		 *   returned to the calling function as is.
+		 */
 		log_debug("waitpid: %s exit status code %d\n",
 		          PPP_DAEMON, exit_status);
-#if HAVE_USR_SBIN_PPPD
-		if (exit_status >= ARRAY_SIZE(pppd_message) || exit_status < 0) {
-			log_error("%s: Returned an unknown exit status: %d\n",
+		if (exit_status >= ARRAY_SIZE(ppp_message)) {
+			log_error("%s: Returned an unknown exit status code: %d\n",
 			          PPP_DAEMON, exit_status);
 		} else {
 			switch (exit_status) {
-			case 0: // success
+			/*
+			 * PPP exit status codes considered as success
+			 */
+			case 0:
 				log_debug("%s: %s\n",
-				          PPP_DAEMON, pppd_message[exit_status]);
-				break;
-			case 16: // emitted when exiting normally
-				log_info("%s: %s\n",
-				         PPP_DAEMON, pppd_message[exit_status]);
+				          PPP_DAEMON, ppp_message[exit_status]);
 				break;
+#if HAVE_USR_SBIN_PPPD
+			case 16: // emitted by Ctrl+C or "kill -15"
+				if (get_sig_received() == SIGINT
+				    || get_sig_received() == SIGTERM) {
+					log_info("%s: %s\n",
+					         PPP_DAEMON, ppp_message[exit_status]);
+					exit_status = 0;
+					break;
+				}
+#endif
+			/*
+			 * PPP exit status codes considered as failure
+			 */
 			default:
-				log_error("%s: %s\n",
-				          PPP_DAEMON, pppd_message[exit_status]);
+				if (ppp_message[exit_status])
+					log_error("%s: %s\n",
+					          PPP_DAEMON, ppp_message[exit_status]);
+				else
+					log_error("%s: Returned an unexpected exit status code: %d\n",
+					          PPP_DAEMON, exit_status);
 				break;
 			}
 		}
-#else
-		// ppp exit codes in the FreeBSD case
-		switch (exit_status) {
-		case 0: // success and EX_NORMAL as defined in ppp source directly
-			log_debug("%s: %s\n", PPP_DAEMON, pppd_message[exit_status]);
-			break;
-		case 1:
-		case 127:
-		case 255: // abnormal exit with hard-coded error codes in ppp
-			log_error("%s: exited with return value of %d\n",
-			          PPP_DAEMON, exit_status);
-			break;
-		default:
-			log_error("%s: %s (%d)\n", PPP_DAEMON, strerror(exit_status),
-			          exit_status);
-			break;
-		}
-#endif
+		return exit_status;
 	} else if (WIFSIGNALED(status)) {
 		int signal_number = WTERMSIG(status);
 
+		/*
+		 * For now we do not consider interruption of the PPP process by
+		 * a signal as a failure. Should we?
+		 */
 		log_debug("waitpid: %s terminated by signal %d\n",
 		          PPP_DAEMON, signal_number);
 		log_error("%s: terminated by signal: %s\n",
@@ -458,19 +512,17 @@ int ppp_interface_is_up(struct tunnel *tunnel)
 	}
 
 	for (ifa = ifap; ifa != NULL; ifa = ifa->ifa_next) {
-		if ((
+		if (
 #if HAVE_USR_SBIN_PPPD
-		            (tunnel->config->pppd_ifname
-		             && strstr(ifa->ifa_name, tunnel->config->pppd_ifname)
-		             != NULL)
-		            || strstr(ifa->ifa_name, "ppp") != NULL
+		        ((tunnel->config->pppd_ifname &&
+		          strstr(ifa->ifa_name, tunnel->config->pppd_ifname) != NULL) ||
+		         strstr(ifa->ifa_name, "ppp") != NULL) &&
 #endif
 #if HAVE_USR_SBIN_PPP
-		            strstr(ifa->ifa_name, "tun") != NULL
+		        strstr(ifa->ifa_name, "tun") != NULL &&
 #endif
-		    ) && ifa->ifa_flags & IFF_UP) {
-			if (&(ifa->ifa_addr->sa_family) != NULL
-			    && ifa->ifa_addr->sa_family == AF_INET) {
+		        ifa->ifa_flags & IFF_UP) {
+			if (ifa->ifa_addr && ifa->ifa_addr->sa_family == AF_INET) {
 				struct in_addr if_ip_addr =
 				        cast_addr(ifa->ifa_addr)->sin_addr;
 
@@ -515,6 +567,17 @@ static int get_gateway_host_ip(struct tunnel *tunnel)
 	return 0;
 }
 
+static int get_socket_option(int sockfd, int level, int optname)
+{
+	int optval;
+	socklen_t optlen = sizeof(optval);
+
+	if (getsockopt(sockfd, level, optname, &optval, &optlen) < 0)
+		return -1;
+	assert(optlen == sizeof(optval));
+	return optval;
+}
+
 /*
  * Establish a regular TCP connection.
  */
@@ -523,7 +586,7 @@ static int tcp_connect(struct tunnel *tunnel)
 	int ret, handle;
 	struct sockaddr_in server;
 	char *env_proxy;
-	const int iface_len = strnlen(tunnel->config->iface_name, IFNAMSIZ);
+	const int iface_len = strnlen(tunnel->config->iface_name, IF_NAMESIZE);
 
 	handle = socket(AF_INET, SOCK_STREAM, 0);
 
@@ -531,7 +594,65 @@ static int tcp_connect(struct tunnel *tunnel)
 		log_error("socket: %s\n", strerror(errno));
 		goto err_socket;
 	}
-	if (iface_len == IFNAMSIZ) {
+
+	/*
+	 * Attempt to find default TCP socket options on different platforms.
+	 */
+#ifdef SO_KEEPALIVE
+	ret = get_socket_option(handle, SOL_SOCKET, SO_KEEPALIVE);
+	if (ret < 0)
+		log_warn("getsockopt: %s: %s\n", "SO_KEEPALIVE", strerror(errno));
+	else
+		log_debug("SO_KEEPALIVE: %s\n", (ret ? "ON" : "OFF"));
+#endif
+#ifdef TCP_KEEPIDLE
+	ret = get_socket_option(handle, IPPROTO_TCP, TCP_KEEPIDLE);
+	if (ret < 0)
+		log_warn("getsockopt: %s: %s\n", "TCP_KEEPIDLE", strerror(errno));
+	else
+		log_debug("TCP_KEEPIDLE: %d\n", ret);
+#endif
+#ifdef TCP_KEEPALIVE
+	ret = get_socket_option(handle, IPPROTO_TCP, TCP_KEEPALIVE);
+	if (ret < 0)
+		log_warn("getsockopt: %s: %s\n", "TCP_KEEPALIVE", strerror(errno));
+	else
+		log_debug("TCP_KEEPALIVE: %d\n", ret);
+#endif
+#ifdef TCP_KEEPINTVL
+	ret = get_socket_option(handle, IPPROTO_TCP, TCP_KEEPINTVL);
+	if (ret < 0)
+		log_warn("getsockopt: %s: %s\n", "TCP_KEEPINTVL", strerror(errno));
+	else
+		log_debug("TCP_KEEPINTVL: %d\n", ret);
+#endif
+#ifdef TCP_KEEPCNT
+	ret = get_socket_option(handle, IPPROTO_TCP, TCP_KEEPCNT);
+	if (ret < 0)
+		log_warn("getsockopt: %s: %s\n", "TCP_KEEPCNT", strerror(errno));
+	else
+		log_debug("TCP_KEEPCNT: %d\n", ret);
+#endif
+#ifdef SO_SNDBUF
+	ret = get_socket_option(handle, SOL_SOCKET, SO_SNDBUF);
+	if (ret < 0)
+#ifndef __APPLE__
+		log_warn("getsockopt: %s: %s\n", "SO_SNDBUF", strerror(errno));
+	else
+#endif
+		log_debug("SO_SNDBUF: %d\n", ret);
+#endif
+#ifdef SO_RCVBUF
+	ret = get_socket_option(handle, SOL_SOCKET, SO_RCVBUF);
+	if (ret < 0)
+#ifndef __APPLE__
+		log_warn("getsockopt: %s: %s\n", "SO_RCVBUF", strerror(errno));
+	else
+#endif
+		log_debug("SO_RCVBUF: %d\n", ret);
+#endif
+
+	if (iface_len == IF_NAMESIZE) {
 		log_error("socket: Too long iface name\n");
 		goto err_post_socket;
 	}
@@ -543,8 +664,8 @@ static int tcp_connect(struct tunnel *tunnel)
 		struct ifreq ifr;
 
 		memset(&ifr, 0, sizeof(ifr));
-		if (strlcpy(ifr.ifr_name, tunnel->config->iface_name, IFNAMSIZ)
-		    >= IFNAMSIZ) {
+		if (strlcpy(ifr.ifr_name, tunnel->config->iface_name, IF_NAMESIZE)
+		    >= IF_NAMESIZE) {
 			log_error("interface name too long\n");
 			goto err_post_socket;
 		}
@@ -800,7 +921,7 @@ static int ssl_verify_cert(struct tunnel *tunnel)
 
 	log_error("Gateway certificate validation failed, and the certificate digest is not in the local whitelist. If you trust it, rerun with:\n");
 	log_error("    --trusted-cert %s\n", digest_str);
-	log_error("or add this line to your config file:\n");
+	log_error("or add this line to your configuration file:\n");
 	log_error("    trusted-cert = %s\n", digest_str);
 	log_error("Gateway certificate:\n");
 	log_error("    subject:\n");
@@ -837,12 +958,49 @@ static void ssl_disconnect(struct tunnel *tunnel)
 
 	SSL_shutdown(tunnel->ssl_handle);
 	SSL_free(tunnel->ssl_handle);
+	tunnel->ssl_handle = NULL;
+
 	SSL_CTX_free(tunnel->ssl_context);
+	tunnel->ssl_context = NULL;
+
 	if (close(tunnel->ssl_socket))
 		log_warn("Could not close ssl socket (%s).\n", strerror(errno));
+	tunnel->ssl_socket = -1;
+}
 
-	tunnel->ssl_handle = NULL;
-	tunnel->ssl_context = NULL;
+/*
+ * Query for the pass phrase used for encrypted PEM structures
+ * (normally only private keys).
+ */
+static int pem_passphrase_cb(char *buf, int size, int rwflag, void *u)
+{
+	struct vpn_config *cfg = (struct vpn_config *)u;
+
+	/* We expect to only read PEM pass phrases, not write them. */
+	if (rwflag == 0) {
+		if (!cfg->pem_passphrase_set) {
+			if (size > PEM_PASSPHRASE_SIZE) {
+				read_password(NULL, NULL,
+				              "Enter PEM pass phrase: ",
+				              cfg->pem_passphrase,
+				              PEM_PASSPHRASE_SIZE + 1);
+				cfg->pem_passphrase_set =  1;
+			} else {
+				log_error("Buffer too small for PEM pass phrase: %d.",
+				          size);
+			}
+		}
+		if (cfg->pem_passphrase_set) {
+			assert(strlen(cfg->pem_passphrase) < size);
+			strncpy(buf, cfg->pem_passphrase, size);
+			buf[size - 1] = '\0';
+			return strlen(buf);
+		}
+	} else {
+		log_error("We refuse to write PEM pass phrases!");
+	}
+
+	return -1;
 }
 
 /*
@@ -854,7 +1012,7 @@ int ssl_connect(struct tunnel *tunnel)
 
 	tunnel->ssl_socket = tcp_connect(tunnel);
 	if (tunnel->ssl_socket == -1)
-		return 1;
+		goto err_tcp_connect;
 
 	// https://wiki.openssl.org/index.php/Library_Initialization
 #if OPENSSL_VERSION_NUMBER < 0x10100000L
@@ -862,13 +1020,15 @@ int ssl_connect(struct tunnel *tunnel)
 	SSL_load_error_strings();
 	// Register the available ciphers and digests
 	SSL_library_init();
-#endif
 
 	tunnel->ssl_context = SSL_CTX_new(SSLv23_client_method());
+#else
+	tunnel->ssl_context = SSL_CTX_new(TLS_client_method());
+#endif
 	if (tunnel->ssl_context == NULL) {
 		log_error("SSL_CTX_new: %s\n",
 		          ERR_error_string(ERR_peek_last_error(), NULL));
-		return 1;
+		goto err_ssl_socket;
 	}
 
 	/* Load the OS default CA files */
@@ -880,7 +1040,7 @@ int ssl_connect(struct tunnel *tunnel)
 		                                   tunnel->config->ca_file, NULL)) {
 			log_error("SSL_CTX_load_verify_locations: %s\n",
 			          ERR_error_string(ERR_peek_last_error(), NULL));
-			return 1;
+			goto err_ssl_context;
 		}
 	}
 
@@ -895,19 +1055,23 @@ int ssl_connect(struct tunnel *tunnel)
 				cipher_list = "HIGH:!aNULL:!kRSA:!PSK:!SRP:!MD5:!RC4";
 			tunnel->config->cipher_list = strdup(cipher_list);
 		} else if (tunnel->config->seclevel_1) {
-			const char *cipher_list = "DEFAULT@SECLEVEL=1";
+			static const char cipher_list[] = "DEFAULT@SECLEVEL=1";
 
 			tunnel->config->cipher_list = strdup(cipher_list);
 		}
 	}
 
+	/* Set the password callback for PEM certificates with encryption. */
+	SSL_CTX_set_default_passwd_cb(tunnel->ssl_context, &pem_passphrase_cb);
+	SSL_CTX_set_default_passwd_cb_userdata(tunnel->ssl_context, tunnel->config);
+
 	if (tunnel->config->cipher_list) {
 		log_debug("Setting cipher list to: %s\n", tunnel->config->cipher_list);
 		if (!SSL_CTX_set_cipher_list(tunnel->ssl_context,
 		                             tunnel->config->cipher_list)) {
 			log_error("SSL_CTX_set_cipher_list failed: %s\n",
 			          ERR_error_string(ERR_peek_last_error(), NULL));
-			return 1;
+			goto err_ssl_context;
 		}
 	}
 
@@ -919,7 +1083,7 @@ int ssl_connect(struct tunnel *tunnel)
 		                                   tunnel->config->min_tls)) {
 			log_error("Cannot set minimum protocol version (%s).\n",
 			          ERR_error_string(ERR_peek_last_error(), NULL));
-			return 1;
+			goto err_ssl_context;
 		}
 	}
 #else
@@ -945,12 +1109,13 @@ int ssl_connect(struct tunnel *tunnel)
 		if ((checkopt & sslctxopt) != sslctxopt) {
 			log_error("SSL_CTX_set_options didn't set opt: %s\n",
 			          ERR_error_string(ERR_peek_last_error(), NULL));
-			return 1;
+			goto err_ssl_context;
 		}
 	}
 #endif
 
 	/* Use engine for PIV if user-cert config starts with pkcs11 URI: */
+#ifndef OPENSSL_NO_ENGINE
 	if (tunnel->config->use_engine > 0) {
 		ENGINE *e;
 
@@ -959,13 +1124,13 @@ int ssl_connect(struct tunnel *tunnel)
 		if (!e) {
 			log_error("Could not load pkcs11 Engine: %s\n",
 			          ERR_error_string(ERR_peek_last_error(), NULL));
-			return 1;
+			goto err_ssl_context;
 		}
 		if (!ENGINE_init(e)) {
 			log_error("Could not init pkcs11 Engine: %s\n",
 			          ERR_error_string(ERR_peek_last_error(), NULL));
 			ENGINE_free(e);
-			return 1;
+			goto err_ssl_context;
 		}
 		if (!ENGINE_set_default_RSA(e))
 			abort();
@@ -981,13 +1146,13 @@ int ssl_connect(struct tunnel *tunnel)
 		if (!ENGINE_ctrl_cmd(e, "LOAD_CERT_CTRL", 0, &parms, NULL, 1)) {
 			log_error("PKCS11 ENGINE_ctrl_cmd: %s\n",
 			          ERR_error_string(ERR_peek_last_error(), NULL));
-			return 1;
+			goto err_ssl_context;
 		}
 
 		if (!SSL_CTX_use_certificate(tunnel->ssl_context, parms.cert)) {
 			log_error("PKCS11 SSL_CTX_use_certificate: %s\n",
 			          ERR_error_string(ERR_peek_last_error(), NULL));
-			return 1;
+			goto err_ssl_context;
 		}
 
 		EVP_PKEY * privkey = ENGINE_load_private_key(
@@ -995,30 +1160,30 @@ int ssl_connect(struct tunnel *tunnel)
 		if (!privkey) {
 			log_error("PKCS11 ENGINE_load_private_key: %s\n",
 			          ERR_error_string(ERR_peek_last_error(), NULL));
-			return 1;
+			goto err_ssl_context;
 		}
 
 		if (!SSL_CTX_use_PrivateKey(tunnel->ssl_context, privkey)) {
 			log_error("PKCS11 SSL_CTX_use_PrivateKey_file: %s\n",
 			          ERR_error_string(ERR_peek_last_error(), NULL));
-			return 1;
+			goto err_ssl_context;
 		}
 
 		if (!SSL_CTX_check_private_key(tunnel->ssl_context)) {
 			log_error("PKCS11 SSL_CTX_check_private_key: %s\n",
 			          ERR_error_string(ERR_peek_last_error(), NULL));
-			return 1;
+			goto err_ssl_context;
 		}
 
 	} else {        /* end PKCS11-engine */
+#endif
 
 		if (tunnel->config->user_cert) {
-			if (!SSL_CTX_use_certificate_file(
-			            tunnel->ssl_context, tunnel->config->user_cert,
-			            SSL_FILETYPE_PEM)) {
-				log_error("SSL_CTX_use_certificate_file: %s\n",
+			if (!SSL_CTX_use_certificate_chain_file(
+			            tunnel->ssl_context, tunnel->config->user_cert)) {
+				log_error("SSL_CTX_use_certificate_chain_file: %s\n",
 				          ERR_error_string(ERR_peek_last_error(), NULL));
-				return 1;
+				goto err_ssl_context;
 			}
 		}
 
@@ -1028,7 +1193,7 @@ int ssl_connect(struct tunnel *tunnel)
 			                                 SSL_FILETYPE_PEM)) {
 				log_error("SSL_CTX_use_PrivateKey_file: %s\n",
 				          ERR_error_string(ERR_peek_last_error(), NULL));
-				return 1;
+				goto err_ssl_context;
 			}
 		}
 
@@ -1036,22 +1201,24 @@ int ssl_connect(struct tunnel *tunnel)
 			if (!SSL_CTX_check_private_key(tunnel->ssl_context)) {
 				log_error("SSL_CTX_check_private_key: %s\n",
 				          ERR_error_string(ERR_peek_last_error(), NULL));
-				return 1;
+				goto err_ssl_context;
 			}
 		}
+#ifndef OPENSSL_NO_ENGINE
 	}
+#endif
 
 	tunnel->ssl_handle = SSL_new(tunnel->ssl_context);
 	if (tunnel->ssl_handle == NULL) {
 		log_error("SSL_new: %s\n",
 		          ERR_error_string(ERR_peek_last_error(), NULL));
-		return 1;
+		goto err_ssl_context;
 	}
 
 	if (!SSL_set_fd(tunnel->ssl_handle, tunnel->ssl_socket)) {
 		log_error("SSL_set_fd: %s\n",
 		          ERR_error_string(ERR_peek_last_error(), NULL));
-		return 1;
+		goto err_ssl_handle;
 	}
 	SSL_set_mode(tunnel->ssl_handle, SSL_MODE_AUTO_RETRY);
 
@@ -1060,18 +1227,32 @@ int ssl_connect(struct tunnel *tunnel)
 		log_error("SSL_connect: %s\n"
 		          "You might want to try --insecure-ssl or specify a different --cipher-list\n",
 		          ERR_error_string(ERR_peek_last_error(), NULL));
-		return 1;
+		goto err_ssl_handle;
 	}
 	SSL_set_mode(tunnel->ssl_handle, SSL_MODE_AUTO_RETRY);
 
 	if (ssl_verify_cert(tunnel))
-		return 1;
+		goto err_ssl_handle;
 
 	// Disable SIGPIPE (occurs when trying to write to an already-closed
 	// socket).
 	signal(SIGPIPE, SIG_IGN);
 
 	return 0;
+
+err_ssl_handle:
+	SSL_shutdown(tunnel->ssl_handle);
+	SSL_free(tunnel->ssl_handle);
+	tunnel->ssl_handle = NULL;
+err_ssl_context:
+	SSL_CTX_free(tunnel->ssl_context);
+	tunnel->ssl_context = NULL;
+err_ssl_socket:
+	if (close(tunnel->ssl_socket))
+		log_warn("Could not close ssl socket (%s).\n", strerror(errno));
+	tunnel->ssl_socket = -1;
+err_tcp_connect:
+	return 1;
 }
 
 int run_tunnel(struct vpn_config *config)
@@ -1080,6 +1261,7 @@ int run_tunnel(struct vpn_config *config)
 	struct tunnel tunnel = {
 		.config = config,
 		.state = STATE_DOWN,
+		.ssl_socket = -1,
 		.ssl_context = NULL,
 		.ssl_handle = NULL,
 		.ipv4.ns1_addr.s_addr = 0,
@@ -1107,7 +1289,7 @@ int run_tunnel(struct vpn_config *config)
 	ret = auth_log_in(&tunnel);
 	if (ret != 1) {
 		log_error("Could not authenticate to gateway. Please check the password, client certificate, etc.\n");
-		log_debug("%s %d\n", err_http_str(ret), ret);
+		log_debug("%s (%d)\n", err_http_str(ret), ret);
 		ret = 1;
 		goto err_tunnel;
 	}
@@ -1152,12 +1334,10 @@ int run_tunnel(struct vpn_config *config)
 	                tunnel.cookie);
 	if (ret != 1) {
 		log_error("Could not start tunnel (%s).\n", err_http_str(ret));
-		ret = 1;
 		goto err_start_tunnel;
 	}
 
 	tunnel.state = STATE_CONNECTING;
-	ret = 0;
 
 	// Step 6: perform io between pppd and the gateway, while tunnel is up
 	log_debug("Starting IO through the tunnel\n");
@@ -1171,7 +1351,7 @@ int run_tunnel(struct vpn_config *config)
 	tunnel.state = STATE_DISCONNECTING;
 
 err_start_tunnel:
-	pppd_terminate(&tunnel);
+	ret = pppd_terminate(&tunnel);
 	log_info("Terminated %s.\n", PPP_DAEMON);
 err_tunnel:
 	log_info("Closed connection to gateway.\n");
@@ -1184,7 +1364,7 @@ err_tunnel:
 		log_info("Logged out.\n");
 	}
 
-	// explicitly free the buffer allocated for split routes of the ipv4 config
+	// explicitly free the buffer allocated for split routes of the ipv4 configuration
 	if (tunnel.ipv4.split_rt != NULL) {
 		free(tunnel.ipv4.split_rt);
 		tunnel.ipv4.split_rt = NULL;
